import json
import os
import sys
from contextlib import contextmanager
from importlib import util
from inspect import getmembers, isfunction

import fire
import pandas as pd
from zeno import (
    DistillReturn,
    ModelReturn,
    ZenoParameters,
    distill,
    metric,
    model,
    zeno,
)
from zeno.api import MetricReturn, ZenoOptions

dfs = {}


@contextmanager
def add_to_path(p):
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def parse_testing_file(test_file):
    # To allow relative imports in test files,
    # add their directory to path temporarily.
    with add_to_path(os.path.dirname(os.path.abspath(test_file))):
        spec = util.spec_from_file_location(str(test_file), test_file)
        test_module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(test_module)  # type: ignore

    functions = []
    for func_name, func in getmembers(test_module):
        if isfunction(func):
            if (
                hasattr(func, "predict_function")
                or hasattr(func, "distill_function")
                or hasattr(func, "metric_function")
                or hasattr(func, "inference_function")
            ):
                functions.append(func)
    return functions


def get_metric_function(metric_name):
    def metric_function(df, ops: ZenoOptions):
        if len(df) == 0:
            return MetricReturn(metric=0.0)
        if (
            df[metric_name].dtype == "object"
            and df[metric_name].value_counts().shape[0] <= 2
        ):
            return MetricReturn(
                metric=df[metric_name].eq(df[metric_name][0]).mul(1).mean()
            )
        return MetricReturn(metric=df[metric_name].mean())

    metric_function.__name__ = metric_name

    return metric(metric_function)


@model
def model_fn(name):
    model_df = dfs[name]

    def mod(df, ops):
        return ModelReturn(model_output=model_df["sampled"].loc[df.index].tolist())

    return mod


@distill
def correct(df, ops: ZenoOptions):
    model_name = [i for i in dfs.keys() if i in ops.output_column]
    model_df = dfs[model_name[0]]
    return DistillReturn(distill_output=model_df["correct"].loc[df.index])


@metric
def avg_correct(df, ops: ZenoOptions):
    return MetricReturn(
        metric=df[ops.distill_columns["correct"]].astype(int).mean() * 100
    )


def read_results_file(data):
    sampling_df = pd.DataFrame(
        [
            {
                "id": d["sample_id"],
                "prompt": d["data"]["prompt"],
                "sampled": d["data"]["sampled"][0],
            }
            for d in data[2:]
            if "type" in d and d["type"] == "sampling"
        ]
    )

    match_df = pd.DataFrame(
        [
            {
                "id": d["sample_id"],
                "correct": d["data"]["correct"],
                "expected": d["data"]["expected"],
            }
            for d in data[2:]
            if "type" in d and d["type"] == "match"
        ]
    )

    metric_names = []
    for d in data[2:]:
        if "type" in d and d["type"] == "metrics":
            metric_names = list(d["data"].keys())
            break

    metrics = []
    for d in data[2:]:
        if "type" in d and d["type"] == "metrics":
            met_obj = {"id": d["sample_id"]}
            for name in metric_names:
                met_obj[name] = d["data"][name]
            metrics.append(met_obj)
    metrics_df = pd.DataFrame(metrics)

    df = sampling_df
    if len(match_df) > 0:
        df = df.join(match_df.set_index("id"), on="id")
    if len(metrics_df) > 0:
        df = df.join(metrics_df.set_index("id"), on="id")

    df.set_index("id", inplace=True, drop=False)
    return df, metric_names


def get_model_name(data):
    name = data[0]["spec"]["completion_fns"][0]
    return name.replace(".", "_")


def generate_zeno_config(
    results_file: str, second_results_file: str = None, functions_file: str = None
) -> ZenoParameters:
    if not os.path.exists(results_file):
        print("ERROR: file '{}' does not exist.".format(results_file))
        sys.exit(1)

    data = []
    with open(results_file) as f:
        for line in f:
            data.append(json.loads(line))

    df, metric_names = read_results_file(data)
    dfs[get_model_name(data)] = df

    models = [get_model_name(data)]
    if second_results_file is not None:
        data2 = []
        with open(second_results_file) as f:
            for line in f:
                data2.append(json.loads(line))

        models.append(get_model_name(data2))
        df2, _ = read_results_file(data2)
        dfs[get_model_name(data2)] = df2

    functions = [model_fn]
    if functions_file is not None:
        functions = functions + parse_testing_file(functions_file)

    base_df_columns = ["id", "prompt"]

    for m in metric_names:
        functions = functions + [get_metric_function(m)]

    if "expected" in df.columns:
        functions = functions + [correct, avg_correct]
        base_df_columns = base_df_columns + ["expected"]

    zeno_config = ZenoParameters(
        metadata=df[base_df_columns],
        models=models,
        functions=functions,
        view="openai-chat",
        data_column="prompt",
        id_column="id",
        cache_path="./.zeno_cache_" + data[0]["spec"]["base_eval"],
        port=8080,
        batch_size=100,
        samples=5,
    )

    if "expected" in df.columns:
        zeno_config.label_column = "expected"

    return zeno_config


def main(
    results_file: str, second_results_file: str = None, functions_file: str = None
):
    """Visualize a result from OpenAI evals using Zeno.

    Args:
            results_file (path): Result .jsonl file from OpenAI evals.
            Often stored in the /tmp/evallogs/ directory.

            second_results_file (path): Second result .jsonl file from OpenAI
            evals for comparison. Often stored in the /tmp/evallogs/ directory.

            functions_file (path, optional): Path to a Python file containing
            additional Zeno processing functions. Defaults to None.
    """

    config = generate_zeno_config(results_file, second_results_file, functions_file)
    zeno(config)


def cli():
    fire.Fire(main)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeno_evals']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.5.0,<0.6.0', 'pandas>=1.5.3,<2.0.0', 'zenoml>=0.4.10,<0.5.0']

entry_points = \
{'console_scripts': ['zeno-evals = zeno_evals.main:cli']}

setup_kwargs = {
    'name': 'zeno-evals',
    'version': '0.1.7',
    'description': 'Visualize OpenAI evals with Zeno',
    'long_description': '# Zeno 🤝 OpenAI Evals\n\nUse [Zeno](https://github.com/zeno-ml/zeno) to visualize the results of [OpenAI Evals](https://github.com/openai/evals/blob/main/docs/eval-templates.md).\n\nhttps://user-images.githubusercontent.com/4563691/225655166-9fd82784-cf35-47c1-8306-96178cdad7c1.mov\n\n_Example using `zeno-evals` to explore the results of an OpenAI eval on multiple choice medicine questions (MedMCQA)_\n\n### Usage\n\n```bash\npip install zeno-evals\n```\n\nRun an evaluation following the [evals instructions](https://github.com/openai/evals/blob/main/docs/run-evals.md). This will produce a cache file in `/tmp/evallogs/`.\n\nPass this file to the `zeno-evals` command:\n\n```bash\nzeno-evals /tmp/evallogs/my_eval_cache.jsonl\n```\n\n### Example\n\nSingle example looking at US tort law questions:\n\n```bash\nzeno-evals ./examples/example.jsonl\n```\n\nAnd an example of comparison between two models:\n\n```bash\nzeno-evals ./examples/crossword-turbo.jsonl ./examples/crossword-turbo-0301.jsonl\n```\n\nAnd lastly, we can pass additional [Zeno functions](https://zenoml.com/docs/api) to provide more context to the results:\n\n```bash\npip install wordfreq\nzeno-evals ./examples/crossword-turbo.jsonl ./examples/crossword-turbo-0301.jsonl --functions_file ./examples/crossword_fns.py\n```\n',
    'author': 'Alex Cabrera',
    'author_email': 'alex.cabrera@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<=3.11',
}


setup(**setup_kwargs)
